package game;

import java.util.*;

/**
 * The map layout. Frozen, and hashable.
 * Can be used as an individual in the evolutionary algorithm.
 */
public class GameLevel {
	public static final int numberOfVertices = 8;
	private MapVertex[] vertices;//do not share references between instances!
	private MapDirectedEdgeType[][] directededges;//do not share references between instances! diagonal entries are null.
	
	public MapVertex getVertex(Location loc) {
		return vertices[loc.arrayIndex()];
	}
	
	public MapDirectedEdgeType getDirectedEdge(Location locFrom, Location locTo) {
		return directededges[locFrom.arrayIndex()][locTo.arrayIndex()];
	}
	
	public GameLevel(GameLevel copy) {
		vertices = new MapVertex[numberOfVertices];
		directededges = new MapDirectedEdgeType[numberOfVertices][numberOfVertices];
		for (int i=0;i<numberOfVertices;i++) {
			vertices[i]=copy.vertices[i];
		}
		for (int i=0;i<numberOfVertices;i++) {
			for (int j=0;j<numberOfVertices;j++) {
				directededges[i][j] = copy.directededges[i][j];
			}
		}
	}
	/**
	 * Initialize with random elements.
	 */
	public GameLevel() {
		vertices = new MapVertex[numberOfVertices];
		directededges = new MapDirectedEdgeType[numberOfVertices][numberOfVertices];
		for (int i=0;i<numberOfVertices;i++) {
			vertices[i]=new MapVertex();
		}
		for (int i=0;i<numberOfVertices;i++) {
			for (int j=0;j<numberOfVertices;j++) {
				if (i==j) {
					directededges[i][j] = null;
					continue;
				}
				if (i<j) {
					if (Rand.rand.nextBoolean()) {
						Vector<MapDirectedEdgeType> pair = MapDirectedEdgeType.randomPair();
						directededges[i][j]= pair.get(0);
						directededges[j][i]= pair.get(1);
					} else {
						Vector<MapDirectedEdgeType> pair = MapDirectedEdgeType.randomPair();
						directededges[i][j]= pair.get(1);
						directededges[j][i]= pair.get(0);
					}
				}
			}
		}
		if (Rand.rand.nextBoolean()) {
			int r = Rand.rand.nextInt(numberOfVertices);
			vertices[r] = vertices[r].setItem(Item.BUTTON1, true);
		}
		if (Rand.rand.nextBoolean()) {
			int r = Rand.rand.nextInt(numberOfVertices);
			vertices[r] = vertices[r].setItem(Item.BUTTON2, true);
		}
		if (Rand.rand.nextBoolean()) {
			int r = Rand.rand.nextInt(numberOfVertices);
			vertices[r] = vertices[r].setItem(Item.BUTTON3, true);
		}
		ensureSingletons();
	}
	/**
	 * make sure that there are no obviously illegal items. For example make sure that there are no 2 buttons 1.
	 * also ensure that there are at most a certain number of portalable fields.
	 */
	private void ensureSingletons() {
		boolean changeOccured=true;
		while (changeOccured) {
			changeOccured=false;
			
			{
				Vector<Location> portalable = new Vector<Location>();
				for (int i=0;i<numberOfVertices;i++) {
					Location loc = new Location(i);
					if (getVertex(loc).isPortalable()) portalable.add(loc);
				}
				while(portalable.size()>5) {
					Location loc = portalable.get(Rand.rand.nextInt(portalable.size()));
					portalable.remove(loc);
					vertices[loc.arrayIndex()]=vertices[loc.arrayIndex()].setPortalability(0);
					changeOccured=true;
				}
			}
			
			{
				nextitem:
				for (Item item : Item.necessary) {
					for (int i=0;i<numberOfVertices;i++) {
						Location loc = new Location(i);
						if (getVertex(loc).hasItem(item)) continue nextitem;
					}
					Location loc = new Location(Rand.rand.nextInt(numberOfVertices));
					vertices[loc.arrayIndex()]=vertices[loc.arrayIndex()].setItem(item, true);
				}
			}
			
			{
				for (Item item : Item.values()) {
					Vector<Location> appearances = new Vector<Location>();
					for (int i=0;i<numberOfVertices;i++) {
						Location loc = new Location(i);
						if (getVertex(loc).hasItem(item)) appearances.add(loc);
					}
					while(appearances.size()>1) {
						Location loc = appearances.get(Rand.rand.nextInt(appearances.size()));
						appearances.remove(loc);
						vertices[loc.arrayIndex()]=vertices[loc.arrayIndex()].setItem(item,false);
						changeOccured=true;
					}
				}
			}
			
			{
				for (int b=1;b<=3;b++) {
					Location buttonpos = null;
					for (int i=0;i<numberOfVertices;i++) {
						Location loc = new Location(i);
						if (b==1 && getVertex(loc).hasItem(Item.BUTTON1)) buttonpos=loc;
						if (b==2 && getVertex(loc).hasItem(Item.BUTTON2)) buttonpos=loc;
						if (b==3 && getVertex(loc).hasItem(Item.BUTTON3)) buttonpos=loc;
					}
					if (buttonpos==null) continue;
					for (int i=0;i<numberOfVertices;i++) {
						Location neigh = new Location(i);
						if (buttonpos.equals(neigh)) continue;
						MapDirectedEdgeType edge = getDirectedEdge(buttonpos, neigh);
						if (edge.buttonInfluence()==b && b==1) vertices[buttonpos.arrayIndex()]=vertices[buttonpos.arrayIndex()].setItem(Item.BUTTON1, false);
						if (edge.buttonInfluence()==b && b==2) vertices[buttonpos.arrayIndex()]=vertices[buttonpos.arrayIndex()].setItem(Item.BUTTON2, false);
						if (edge.buttonInfluence()==b && b==3) vertices[buttonpos.arrayIndex()]=vertices[buttonpos.arrayIndex()].setItem(Item.BUTTON3, false);
					}
				}
			}
		}
	}
	
	private double cleanFitness() {
		double ret = PuzzleSolver.calculateFitness(this, null);
		if (ret==0) return Double.NEGATIVE_INFINITY;
		for (int i=0;i<numberOfVertices;i++) {
			ret-=vertices[i].complexity()*0.1d;
		}
		for (int i=0;i<numberOfVertices;i++) {
			for (int j=0;j<numberOfVertices;j++) {
				if (i==j) continue;
				ret-=directededges[i][j].complexity()*0.10d;
			}
		}
		return ret;
	}
	public GameLevel clean() {
		GameLevel ret = new GameLevel(this);
		boolean changeOccured=true;
		while (changeOccured) {
			System.out.println("starting a new round of postprocessing. This can take a few hours.");
			changeOccured=false;
			{
				Vector<Location> allLocations = new Vector<Location>();
				for (int i=0;i<numberOfVertices;i++) {
					Location loc = new Location(i);
					allLocations.add(loc);
				}
				while (!allLocations.isEmpty()) {
					System.out.println((1+numberOfVertices-allLocations.size())+"/"+(numberOfVertices+numberOfVertices*(numberOfVertices-1)));
					Location loc = allLocations.get(Rand.rand.nextInt(allLocations.size()));
//					System.out.println("\nloc="+loc);
					allLocations.remove(loc);
					double originalfitness = ret.cleanFitness();
//					System.out.println("origfit="+originalfitness);
					double maxfitness = originalfitness;
					GameLevel originalLevel = new GameLevel(ret);
					GameLevel bestLevel = new GameLevel(ret);
					for (int i=0;i<MapVertex.getNumberOfPossibilities();i++) {
						MapVertex v = MapVertex.getPossibility(i);
						v = v.setItem(Item.CUBEDROPPER1, originalLevel.getVertex(loc).isCubeDropper(1));
						v = v.setItem(Item.CUBEDROPPER2, originalLevel.getVertex(loc).isCubeDropper(2));
						v = v.setItem(Item.CUBESPAWNBUTTON1, originalLevel.getVertex(loc).isCubeSpawnButton(1));
						v = v.setItem(Item.CUBESPAWNBUTTON2, originalLevel.getVertex(loc).isCubeDropper(2));
						ret.vertices[loc.arrayIndex()]=v;
						ret.ensureSingletons();
						double fitness = ret.cleanFitness();
//						System.out.println("cleanfitness="+fitness+"  maxf="+maxfitness+"  ch="+changeOccured);
						if (fitness>maxfitness) {
							maxfitness=fitness;
							bestLevel = new GameLevel(ret);
							changeOccured = true;
						}
//						else if (fitness==maxfitness) {System.out.print("=");}
//						else if (fitness<maxfitness) {System.out.print("<");}
						ret = new GameLevel(originalLevel);
					}
					if (maxfitness>originalfitness) {
						ret = new GameLevel(bestLevel);
					}
				}
			}
//			System.out.println("preliminary fitness = "+ret.cleanFitness());
			{
				Vector<Vector<Location>> allUndirEdges = new Vector<Vector<Location>>();
				for (int i=0;i<numberOfVertices;i++) {
					for (int j=0;j<numberOfVertices;j++) {
						if (i==j) continue;
						Vector<Location> undirEdge = new Vector<Location>();
						undirEdge.add(new Location(i));
						undirEdge.add(new Location(j));
						allUndirEdges.add(undirEdge);
					}
				}
				while (!allUndirEdges.isEmpty()) {
					System.out.println(((numberOfVertices*(numberOfVertices-1))-allUndirEdges.size()+numberOfVertices+1)+"/"+(numberOfVertices+numberOfVertices*(numberOfVertices-1)));
					Vector<Location> undirEdge = allUndirEdges.get(Rand.rand.nextInt(allUndirEdges.size()));
//					System.out.println(ret.humanReadable());
//					System.out.println("\nundirEdge="+undirEdge);
					allUndirEdges.remove(undirEdge);
					double originalfitness = ret.cleanFitness();
//					System.out.println("origfit="+originalfitness);
					double maxfitness = originalfitness;
					GameLevel originalLevel = new GameLevel(ret);
					GameLevel bestLevel = null;

					for (int index=0; index<=MapDirectedEdgeType.maxIndex(); index++) {
						Vector<MapDirectedEdgeType> pair = MapDirectedEdgeType.pair(index);
						ret.directededges[undirEdge.get(0).arrayIndex()][undirEdge.get(1).arrayIndex()]=pair.get(0);
						ret.directededges[undirEdge.get(1).arrayIndex()][undirEdge.get(0).arrayIndex()]=pair.get(1);
						double fitness = ret.cleanFitness();
						if (fitness>maxfitness) {
							maxfitness=fitness;
							bestLevel=new GameLevel(ret);
						}
//						else if (fitness==maxfitness) System.out.print("=");
//						else if (fitness<maxfitness) System.out.print("<");
						ret = new GameLevel(originalLevel);
					}
					if (maxfitness>originalfitness) {
						ret = new GameLevel(bestLevel);
						changeOccured=true;
					}
				}
			}
			System.out.println("finished this round of postprocessing.");
		}
		System.out.println("no more rounds of postprocessing.");
		return ret;
	}
	
	
	/**
	 * returns a new GameLevel object with the same complexity that does not contain unnecessary Fizzlers, Fires, Open Paths, Buttons, but contains more Portal Surfaces
	 */
	public GameLevel mutate() {
		GameLevel ret = new GameLevel();
		for (int i=0;i<numberOfVertices;i++) {
			if (Rand.rand.nextInt(100)!=0) {
				ret.vertices[i]=this.vertices[i];
			}
		}
		for (int i=0;i<numberOfVertices;i++) {
			for (int j=0;j<numberOfVertices;j++) {
				if (Rand.rand.nextInt(100)!=0) {
					Vector<MapDirectedEdgeType> pair = MapDirectedEdgeType.randomPair();
					ret.directededges[i][j]= pair.get(0);
					ret.directededges[j][i]= pair.get(1);
				}
			}
		}
		ret.ensureSingletons();
		return ret;
	}
	public static GameLevel recombine(GameLevel parent1, GameLevel parent2) {
		GameLevel ret = new GameLevel();
		for (int i=0;i<numberOfVertices;i++) {
			ret.vertices[i]=Rand.rand.nextBoolean()?parent1.vertices[i]:parent2.vertices[i];
		}
		for (int i=0;i<numberOfVertices;i++) {
			for (int j=i;j<numberOfVertices;j++) {
				boolean b = Rand.rand.nextBoolean();
				ret.directededges[i][j] = b?parent1.directededges[i][j]:parent2.directededges[i][j];
				ret.directededges[j][i] = b?parent1.directededges[j][i]:parent2.directededges[j][i];
			}
		}
		ret.ensureSingletons();
		return ret;
	}
//	public GameLevel(String s) {
//		String[] tokens = s.split(" ");
//		int c=0;
//		vertices = new MapVertex[numberOfVertices];
//		directededges = new MapDirectedEdge[numberOfVertices][numberOfVertices];
//		for (int i=0;i<numberOfVertices;i++) {
//			vertices[i] = new MapVertex(tokens[c]);
//			c++;
//		}
//		for (int i=0;i<numberOfVertices;i++) {
//			for (int j=0;j<numberOfVertices;j++) {
//				if(i!=j) {
//					directededges[i][j] = new MapDirectedEdge(tokens[c],new Location(i),new Location(j));
//					c++;
//				}
//			}
//		}
//	}
//	public GameLevel(String s) {
//		String[] tokens = s.split(" ");
//		int c=0;
//		vertices = new MapVertex[numberOfVertices];
//		directededges = new MapDirectedEdgeType[numberOfVertices][numberOfVertices];
//		for (int i=0;i<numberOfVertices;i++) {
//			vertices[i] = new MapVertex(tokens[c]);
//			c++;
//		}
//		for (int i=0;i<numberOfVertices;i++) {
//			for (int j=0;j<numberOfVertices;j++) {
//				if(i!=j) {
//					directededges[i][j] = MapDirectedEdgeType.valueOf(tokens[c]);
//					c++;
//				}
//			}
//		}
//		ensureSingletons();
//	}
	public GameLevel(String s) {
		String[] tokens = s.split("\n");
		int c=0;
		vertices = new MapVertex[numberOfVertices];
		directededges = new MapDirectedEdgeType[numberOfVertices][numberOfVertices];
		for (int i=0;i<numberOfVertices;i++) {
			int[] vertex = new int[]{0,0,0,0,0,0,0,0};
			if (tokens[c].contains("[portal surface on wall]")) vertex[0]=2;
			if (tokens[c].contains("[portal surface on ceiling]")) vertex[0]=1;
			if (tokens[c].contains("[storage cube dropper]")) vertex[1]=1;
			if (tokens[c].contains("[companion cube dropper]")) vertex[2]=1;
			if (tokens[c].contains("[button to spawn storage cube]")) vertex[3]=1;
			if (tokens[c].contains("[button to spawn companion cube]")) vertex[4]=1;
			if (tokens[c].contains("[cube button that activates things with label '1']")) vertex[5]=1;
			if (tokens[c].contains("[cube button that activates things with label '6']")) vertex[6]=1;
			if (tokens[c].contains("[cube button that activates things with label '7']")) vertex[7]=1;
			String token = ""+vertex[0]+vertex[1]+vertex[2]+vertex[3]+vertex[4]+vertex[5]+vertex[6]+vertex[7];
			vertices[i] = new MapVertex(token);
			c++;
		}
		c++;//"descriptions of the connections between the locations:"
		int d=c;
		for (int i=0;i<numberOfVertices;i++) {
			for (int j=0;j<numberOfVertices;j++) {
				if(i!=j) {
					directededges[i][j]=MapDirectedEdgeType.WALL;
				}
			}
		}
		for (int i=0;i<numberOfVertices;i++) {
			for (int j=0;j<numberOfVertices;j++) {
				if(i!=j) {
					int from = new Integer(tokens[d].substring(5,6));
					int to = new Integer(tokens[d].substring(10,11));
					if (i!=from || j!=to) {
						c++;continue;
					}
					MapDirectedEdgeType found = null;
					for (MapDirectedEdgeType mdet : MapDirectedEdgeType.values()) {
						if (tokens[d].substring(13).equals(mdet.humanReadable())) {
							found = mdet;
						}
					}
					if (found==null) throw new Error();
					for (int pairindex = 0; pairindex<=MapDirectedEdgeType.maxIndex(); pairindex++) {
						Vector<MapDirectedEdgeType> pair = MapDirectedEdgeType.pair(pairindex);
						if (pair.firstElement().equals(found)) {
							directededges[i][j] = found;
							directededges[j][i] = pair.lastElement();
						}
					}
					c++;d++;
				}
			}
		}
		ensureSingletons();
	}
	@Override
	public String toString() {
		String ret="";
		for (int i=0;i<numberOfVertices;i++) {
			ret+=vertices[i].toString()+" ";
		}
		for (int i=0;i<numberOfVertices;i++) {
			for (int j=0;j<numberOfVertices;j++) {
				if(i!=j) {
					ret+=directededges[i][j].toString()+" ";
				}
			}
		}
		return ret;
	}
	@Override
	public int hashCode() {
		int ret = 1;
		for (int i=0;i<numberOfVertices;i++) {
			ret+=vertices[i].hashCode();
			ret*=13;
		}
		for (int i=0;i<numberOfVertices;i++) {
			ret*=17;
			for (int j=0;j<numberOfVertices;j++) {
				if (i==j) continue;
				ret+=directededges[i][j].hashCode();
				ret*=5;
			}
		}
		return ret;
	}
	@Override
	public boolean equals(Object o) {
		if (o instanceof GameLevel) {
			GameLevel g = (GameLevel)o;
			for (int i=0;i<numberOfVertices;i++) {
				if (!g.vertices[i].equals(vertices[i])) return false;
			}
			for (int i=0;i<numberOfVertices;i++) {
				for (int j=0;j<numberOfVertices;j++) {
					if (i==j) continue;
					if (!g.directededges[i][j].equals(directededges[i][j])) return false;
				}
			}
		}
		return false;
	}
	
//	private String s(int i, String s) {
//		if (i==0) return " ";
//		if (i==4) return s;
//		return ""+i;
//	}
	public String humanReadable() {
		String ret = "";
		ret+="descriptions of the locations 0 to "+(numberOfVertices-1)+":"+System.getProperty("line.separator");
		for (int i=0;i<numberOfVertices;i++) {
			ret+=i+":"+vertices[i].humanReadable(i)+System.getProperty("line.separator");
		}
		ret+="descriptions of the connections between the locations:"+System.getProperty("line.separator");
		for (int i=0;i<numberOfVertices;i++) {
			for (int j=0;j<numberOfVertices;j++) {
				if (i==j) continue;
				if (!directededges[i][j].isJustCompanion()) {
					ret+="from "+i+" to "+j+": "+directededges[i][j].humanReadable()+System.getProperty("line.separator");
				}
			}
		}
		ret+="You can use a graph visualization software to visualize the chamber's layout, for example GraphViz."+System.getProperty("line.separator");
		return ret;
	}
}
