package game;

import java.util.*;

/**
 * The state of a game. Frozen and hashable.
 * This is a vertex in the fine-grained gamestate graph.
 * The portal positions can be NOTTHERE.
 * 
 * Comparable because a strongly connected component should have a representative (the normalized game state)
 */
public class GameState implements Comparable<GameState> {
	protected GameLevel gamelevel;
	protected Location playerpos;
	protected Vector<Location> cubepos;//do not share references between instances!
	protected Vector<Location> portalpos;//do not share references between instances!
	public GameState(GameState g) {
		gamelevel = g.gamelevel;
		playerpos = g.playerpos;
		cubepos = new Vector<Location>();cubepos.addAll(g.cubepos);
		portalpos = new Vector<Location>();portalpos.addAll(g.portalpos);
	}
	public GameState(GameLevel gamelevel) {
		this.gamelevel = gamelevel;
		portalpos = new Vector<Location>();
		portalpos.add(Location.NOTTHERE);
		portalpos.add(Location.NOTTHERE);
		cubepos = new Vector<Location>();
		cubepos.add(Location.NOTTHERE);
		cubepos.add(Location.NOTTHERE);
		playerpos = Location.STARTPOS;
	}
	@Override
	public int compareTo(GameState g) {
		int ret;
		ret = playerpos.compareTo(g.playerpos);
		if (ret!=0) return ret;
		ret = cubepos.get(0).compareTo(g.cubepos.get(0));
		if (ret!=0) return ret;
		ret = cubepos.get(1).compareTo(g.cubepos.get(1));
		if (ret!=0) return ret;
		ret = portalpos.get(0).compareTo(g.portalpos.get(0));
		if (ret!=0) return ret;
		ret = portalpos.get(1).compareTo(g.portalpos.get(1));
		if (ret!=0) return ret;
		return ret;
	}
	public Vector<PlayerAction> generatePlayerActions(boolean cautious) {
		Vector<PlayerAction> ret = new Vector<PlayerAction>();
		if (cautious) return ret;
		//<buttons>
		boolean button1=false;
		boolean button2=false;
		boolean button3=false;
		if (!cautious) {
			for (int b=1;b<=3;b++) {
				for (int i=0;i<GameLevel.numberOfVertices;i++) {
					Location loc = new Location(i);
					if (gamelevel.getVertex(loc).isButton(b)) {
						if (playerpos.equals(loc) || cubepos.get(0).equals(loc) || cubepos.get(1).equals(loc)) {
							if (b==1) button1=true;
							if (b==2) button2=true;
							if (b==3) button3=true;
						}
					}
				}
			}
		}
		//</buttons>
		//<forget portals>
//		if (!portalpos.get(0).equals(Location.NOTTHERE) || !portalpos.get(1).equals(Location.NOTTHERE)) {
//			GameState newg = new GameState(this);
//			newg.portalpos.set(0,Location.NOTTHERE);
//			newg.portalpos.set(1,Location.NOTTHERE);
//			ret.add(new PlayerAction(PlayerActionType.FORGETPORTALS,this,newg));
//		}
		//</forget portals>
		//<place portals>
		{
			Vector<Location> portalpositions = new Vector<Location>();
			if (gamelevel.getVertex(playerpos).isPortalable()) portalpositions.add(playerpos);
			for (int i=0;i<GameLevel.numberOfVertices;i++) {
				Location loc = new Location(i);
				if (!loc.equals(playerpos)) {
					if (gamelevel.getVertex(loc).isPortalable()) {
						if (gamelevel.getDirectedEdge(playerpos, loc).canShoot(button1, button2, button3)) {
							portalpositions.add(loc);
						}
					}
				}
			}
			for (Location loc : portalpositions) {
				GameState newg = new GameState(this);
				newg.portalpos.set(0,loc);
				ret.add(new PlayerAction(PlayerActionType.PLACEPRIMARYPORTAL,this,newg));
				newg = new GameState(this);
				newg.portalpos.set(1,loc);
				ret.add(new PlayerAction(PlayerActionType.PLACESECONDARYPORTAL,this,newg));
			}
		}
		//</place portals>
		//<move and carry and throw, also through portals>
		{
			Vector<LocationPair> walkableEdgesMaybeFizzling = new Vector<LocationPair>();
			HashSet<LocationPair> fizzlingEdges = new HashSet<LocationPair>();
			Vector<LocationPair> cubeThrowEdges = new Vector<LocationPair>();
			for (int i=0;i<GameLevel.numberOfVertices;i++) {
				Location loc = new Location(i);
				if (!loc.equals(playerpos)) {
					if (gamelevel.getDirectedEdge(playerpos, loc).canWalk(button1, button2, button3)>=1) {
						walkableEdgesMaybeFizzling.add(new LocationPair(playerpos, loc));
						if (gamelevel.getDirectedEdge(playerpos, loc).canWalk(button1, button2, button3)==1) {
							fizzlingEdges.add(new LocationPair(playerpos, loc));
						}
					}
					if (gamelevel.getDirectedEdge(playerpos, loc).canThrowCube(button1, button2, button3)) {
						cubeThrowEdges.add(new LocationPair(playerpos, loc));
					}
				}
			}

			//<walk not through a portal>
			for (int i=0;i<walkableEdgesMaybeFizzling.size();i++) {
				LocationPair edge = walkableEdgesMaybeFizzling.get(i);
				Location loc = edge.getTo();
				{
					GameState newg = new GameState(this);
					newg.playerpos=loc;
					if (fizzlingEdges.contains(edge)) {
						newg.portalpos.set(0, Location.NOTTHERE);
						newg.portalpos.set(1, Location.NOTTHERE);
					}
					ret.add(new PlayerAction(PlayerActionType.MOVE,this,newg));
				}
			}
			//</walk not through a portal>
			//<carry not through a portal>
			for (int i=0;i<walkableEdgesMaybeFizzling.size();i++) {
				for (int c=0;c<2;c++) {
					if (cubepos.get(c).equals(playerpos)) {
						LocationPair edge = walkableEdgesMaybeFizzling.get(i);
						Location loc = edge.getTo();
						{
							GameState newg = new GameState(this);
							newg.playerpos=loc;
							newg.cubepos.set(c, loc);
							if (fizzlingEdges.contains(edge)) continue;
							ret.add(new PlayerAction(PlayerActionType.CARRY,this,newg));
						}
					}
				}
				if (cubepos.get(0).equals(playerpos) && cubepos.get(1).equals(playerpos)) {
					LocationPair edge = walkableEdgesMaybeFizzling.get(i);
					Location loc = edge.getTo();
					{
						GameState newg = new GameState(this);
						newg.playerpos=loc;
						newg.cubepos.set(0, loc);
						newg.cubepos.set(1, loc);
						if (fizzlingEdges.contains(edge)) continue;
						ret.add(new PlayerAction(PlayerActionType.CARRY,this,newg));
					}
				}
			}
			//</carry not through a portal>
			//<throw not through a portal>
			for (int i=0;i<cubeThrowEdges.size();i++) {
				for (int c=0;c<2;c++) {
					if (cubepos.get(c).equals(playerpos)) {
						LocationPair edge = cubeThrowEdges.get(i);
						Location loc = edge.getTo();
						{
							GameState newg = new GameState(this);
							newg.cubepos.set(c, loc);
							ret.add(new PlayerAction(PlayerActionType.THROWCUBE,this,newg));
						}
					}
				}
				if (cubepos.get(0).equals(playerpos) && cubepos.get(1).equals(playerpos)) {
					LocationPair edge = cubeThrowEdges.get(i);
					Location loc = edge.getTo();
					{
						GameState newg = new GameState(this);
						newg.cubepos.set(0, loc);
						newg.cubepos.set(1, loc);
						ret.add(new PlayerAction(PlayerActionType.THROWCUBE,this,newg));
					}
				}
			}
			//</throw not through a portal>
			//<walk through a portal>
			if (portalpos.get(0).equals(playerpos) && gamelevel.getVertex(playerpos).isPortalAccessible() && !portalpos.get(1).equals(Location.NOTTHERE) && !portalpos.get(1).equals(playerpos)) {
				Location loc = portalpos.get(1);
				GameState newg = new GameState(this);
				newg.playerpos=loc;
				ret.add(new PlayerAction(PlayerActionType.MOVETHROUGHPRIMARYPORTAL,this,newg));
			}
			if (portalpos.get(1).equals(playerpos) && gamelevel.getVertex(playerpos).isPortalAccessible() && !portalpos.get(0).equals(Location.NOTTHERE) && !portalpos.get(0).equals(playerpos)) {
				Location loc = portalpos.get(0);
				GameState newg = new GameState(this);
				newg.playerpos=loc;
				ret.add(new PlayerAction(PlayerActionType.MOVETHROUGHSECONDARYPORTAL,this,newg));
			}
			//</walk through a portal>
			//<carry through a portal>
			if (portalpos.get(0).equals(playerpos) && gamelevel.getVertex(playerpos).isPortalAccessible() && !portalpos.get(1).equals(Location.NOTTHERE) && !portalpos.get(1).equals(playerpos)) {
				for (int c=0;c<2;c++) {
					if (cubepos.get(c).equals(playerpos)) {
						Location loc = portalpos.get(1);
						GameState newg = new GameState(this);
						newg.playerpos=loc;
						newg.cubepos.set(c, loc);
						ret.add(new PlayerAction(PlayerActionType.CARRYTHROUGHPRIMARYPORTAL,this,newg));
					}
				}
				if (cubepos.get(0).equals(playerpos) && cubepos.get(1).equals(playerpos)) {
					Location loc = portalpos.get(1);
					GameState newg = new GameState(this);
					newg.playerpos=loc;
					newg.cubepos.set(0,loc);
					newg.cubepos.set(1,loc);
					ret.add(new PlayerAction(PlayerActionType.CARRYTHROUGHPRIMARYPORTAL,this,newg));
				}
			}
			if (portalpos.get(1).equals(playerpos) && gamelevel.getVertex(playerpos).isPortalAccessible() && !portalpos.get(0).equals(Location.NOTTHERE) && !portalpos.get(0).equals(playerpos)) {
				for (int c=0;c<2;c++) {
					if (cubepos.get(c).equals(playerpos)) {
						Location loc = portalpos.get(0);
						GameState newg = new GameState(this);
						newg.playerpos=loc;
						newg.cubepos.set(c, loc);
						ret.add(new PlayerAction(PlayerActionType.CARRYTHROUGHSECONDARYPORTAL,this,newg));
					}
				}
				if (cubepos.get(0).equals(playerpos) && cubepos.get(1).equals(playerpos)) {
					Location loc = portalpos.get(0);
					GameState newg = new GameState(this);
					newg.playerpos=loc;
					newg.cubepos.set(0,loc);
					newg.cubepos.set(1,loc);
					ret.add(new PlayerAction(PlayerActionType.CARRYTHROUGHSECONDARYPORTAL,this,newg));
				}
			}
			//</carry through a portal>
			//<throw through a portal>
			if (portalpos.get(0).equals(playerpos) && gamelevel.getVertex(playerpos).isPortalAccessible() && !portalpos.get(1).equals(Location.NOTTHERE) && !portalpos.get(1).equals(playerpos)) {
				for (int c=0;c<2;c++) {
					if (cubepos.get(c).equals(playerpos)) {
						Location loc = portalpos.get(1);
						GameState newg = new GameState(this);
						newg.cubepos.set(c, loc);
						ret.add(new PlayerAction(PlayerActionType.THROWCUBETHROUGHPRIMARYPORTAL,this,newg));
					}
				}
				if (cubepos.get(0).equals(playerpos) && cubepos.get(1).equals(playerpos)) {
					Location loc = portalpos.get(1);
					GameState newg = new GameState(this);
					newg.cubepos.set(0, loc);
					newg.cubepos.set(1, loc);
					ret.add(new PlayerAction(PlayerActionType.THROWCUBETHROUGHPRIMARYPORTAL,this,newg));
				}
			}
			if (portalpos.get(1).equals(playerpos) && gamelevel.getVertex(playerpos).isPortalAccessible() && !portalpos.get(0).equals(Location.NOTTHERE) && !portalpos.get(0).equals(playerpos)) {
				for (int c=0;c<2;c++) {
					if (cubepos.get(c).equals(playerpos)) {
						Location loc = portalpos.get(0);
						GameState newg = new GameState(this);
						newg.cubepos.set(c, loc);
						ret.add(new PlayerAction(PlayerActionType.THROWCUBETHROUGHSECONDARYPORTAL,this,newg));
					}
				}
				if (cubepos.get(0).equals(playerpos) && cubepos.get(1).equals(playerpos)) {
					Location loc = portalpos.get(0);
					GameState newg = new GameState(this);
					newg.cubepos.set(0, loc);
					newg.cubepos.set(1, loc);
					ret.add(new PlayerAction(PlayerActionType.THROWCUBETHROUGHSECONDARYPORTAL,this,newg));
				}
			}
			//</throw through a portal>
		}
		//</move and carry and throw, also through portals>
		//<spawn a new cube>
		for (int j=1;j<=2;j++) {
			if (gamelevel.getVertex(playerpos).isCubeSpawnButton(j)) {
				int pos = Location.NOTTHERE.arrayIndex();
				for (int i=0;i<GameLevel.numberOfVertices;i++) {
					Location loc = new Location(i);
					if (gamelevel.getVertex(loc).isCubeDropper(j)) pos = i;
				}
				GameState newg = new GameState(this);
				newg.cubepos.set(j-1,new Location(pos));
				ret.add(new PlayerAction(PlayerActionType.CLICK,this,newg));
			}
		}
		//</spawn a new cube>
		return ret;
	}
	public boolean hasWon() {
		return playerpos.equals(Location.ENDPOS) && cubepos.get(0).equals(Location.ENDPOS) && cubepos.get(1).equals(Location.ENDPOS);
	}
	/**
	 * @return the connected component and its normalized game state
	 */
	private NormalizedGameState generateStronglyConnectedComponent(Long killtime) throws TimeUpException {
		GameState winner = this;
		HashSet<GameState> expanded = new HashSet<GameState>();
		HashSet<GameState> border = new HashSet<GameState>();
		border.add(this);
		while(!border.isEmpty()) {
			if (killtime!=null && TimerThread.currentTimeMillis()>=killtime) throw new TimeUpException();
			GameState expand = border.iterator().next();
			border.remove(expand);
			expanded.add(expand);
			if (expand.compareTo(winner)>0) winner=expand;
			Vector<PlayerAction> actions = expand.generatePlayerActions(true);
			for (PlayerAction action : actions) {
				Vector<PlayerAction> backactions = action.getNewGamestate().generatePlayerActions(true);
				boolean canGoBack = false;
				for (PlayerAction backaction : backactions) {
					if (backaction.getNewGamestate().equals(expand)) {
						canGoBack=true;
						break;
					}
				}
				if (canGoBack) {
					if (!expanded.contains(action.getNewGamestate())) {
						border.add(action.getNewGamestate());
					}
				}
			}
		}
		return new NormalizedGameState(winner,expanded);
	}
	
	private static GameLevel gameLevelForCache = null;
	private static HashMap<GameState,NormalizedGameState> cachedNormalizedGameStates = null;
	/**
	 * supports caching
	 */
	public NormalizedGameState normalize(Long killtime) throws TimeUpException {
		if (gameLevelForCache != gamelevel) {
			cachedNormalizedGameStates = new HashMap<GameState,NormalizedGameState>();
			gameLevelForCache = gamelevel;
		}
		if (cachedNormalizedGameStates.containsKey(this)) return cachedNormalizedGameStates.get(this);
		
		NormalizedGameState ret = generateStronglyConnectedComponent(killtime);
		cachedNormalizedGameStates.put(this, ret);
		return ret;
	}
	public NormalizedGameState normalizeArbitraryTime() {
		NormalizedGameState ret = null;
		try {
			ret = normalize(null);
		} catch (TimeUpException e) {}
		return ret;
	}
	@Override
	public boolean equals(Object o) {
		if (o instanceof GameState) {
			GameState gs = (GameState)o;
			if (gs.gamelevel!=gamelevel) return false;
			return playerpos.equals(gs.playerpos) && portalpos.equals(gs.portalpos) && cubepos.equals(gs.cubepos);
		}
		return false;
	}

	@Override
	public int hashCode() {
        int hash = 1;
        hash = hash * 17 + playerpos.hashCode();
        hash = hash * 31 + portalpos.hashCode();
        hash = hash * 13 + cubepos.hashCode();
        return hash;
	}
	
	@Override
	public String toString() {
		return playerpos+" "+cubepos.get(0)+" "+cubepos.get(1)+" "+portalpos.get(0)+" "+portalpos.get(1);
	}
	public String humanReadable() {
		return "{player at "+playerpos
				+(cubepos.get(0).equals(Location.NOTTHERE)?"":", storage cube at "+cubepos.get(0))
				+(cubepos.get(1).equals(Location.NOTTHERE)?"":", companion cube at "+cubepos.get(1))
				+(portalpos.get(0).equals(Location.NOTTHERE)?"":", blue portal at "+portalpos.get(0))
				+(portalpos.get(1).equals(Location.NOTTHERE)?"":", orange portal at "+portalpos.get(1))
				+"}";
	}
	
	public GameState debugteleport(Location position) {
		GameState ret = new GameState(this);
		ret.playerpos= position;
		return ret;
	}

}
