package game;

import java.util.*;

public class PuzzleSolver {
	private static HashMap<GameLevel,Double> fitnesscache = new HashMap<GameLevel,Double>();
	public static double calculateFitness(GameLevel gamelevel, Long killtime) {
		try {
			if (fitnesscache.containsKey(gamelevel)) {
				return fitnesscache.get(gamelevel);
			}
			PuzzleSolver solver = new PuzzleSolver(gamelevel, killtime);
			Vector<PlayerAction> solution = solver.solve(killtime);
			if (solution==null) {
				fitnesscache.put(gamelevel,0d);
				return 0;
			}
			double distance = 0;
			for (PlayerAction action : solution) {
				distance+=action.getCost();
			}
			return distance;
		} catch (TimeUpException e) {/*System.out.print("#");*/return 0;}
	}
	
	
	protected HashSet<NormalizedGameState> expanded;
	protected Vector<NormalizedGameState> border;//ordered!
	protected HashMap<NormalizedGameState,Double> distancetostart;
	protected HashMap<NormalizedGameState,Vector<PlayerAction>> predecessoraction;//first element in vector is the best predecessoraction

	
	GameLevel gamelevel;
	protected NormalizedGameState startstate;
	public PuzzleSolver(GameLevel gamelevel, Long killtime) throws TimeUpException {
		this.gamelevel = gamelevel;
		expanded = new HashSet<NormalizedGameState>();
		border = new Vector<NormalizedGameState>();
		predecessoraction = new HashMap<NormalizedGameState,Vector<PlayerAction>>();//the action that moves from predecessor to this state
		distancetostart = new HashMap<NormalizedGameState,Double>();
		startstate = new GameState(gamelevel).normalize(killtime);
		border.add(startstate);
		distancetostart.put(startstate, 0d);
	}
	
	

	public Vector<PlayerAction> solve(Long killtime) throws TimeUpException {
		Comparator<NormalizedGameState> comp = new Comparator<NormalizedGameState>(){
			@Override
			public int compare(NormalizedGameState g1, NormalizedGameState g2) {
				double d1 = distancetostart.get(g1);
				double d2 = distancetostart.get(g2);
				return (d1==d2?g1.getRepresentativeGameState().compareTo(g2.getRepresentativeGameState()):(d1<d2?-1:1));
			}};
		HashSet<NormalizedGameState> winningstates = new HashSet<NormalizedGameState>();
		Vector<PlayerAction> solution = null;
		while (!border.isEmpty()) {
			if (killtime!=null && TimerThread.currentTimeMillis()>=killtime) throw new TimeUpException();
			NormalizedGameState expand = border.firstElement();
			if (expand.hasWon()) {
				//BEST SOLUTION FOUND!
				winningstates.add(expand);
				if (solution==null) {
					solution = new Vector<PlayerAction>();
					PlayerAction monkey = predecessoraction.get(expand).firstElement();
					if (monkey==null) solution = new Vector<PlayerAction>();//trivial 1-step solution
					else {
						while (!monkey.getOldGamestate().normalize(killtime).equals(startstate)) {
							solution.add(monkey);
							monkey = predecessoraction.get(monkey.getOldGamestate().normalize(killtime)).firstElement();
						}
						solution.add(predecessoraction.get(monkey.getNewGamestate().normalize(killtime)).firstElement());
						Collections.reverse(solution);
					}
				}
			}
			Vector<PlayerAction> actions = expand.generatePlayerActions(killtime);
			for (PlayerAction action : actions) {
				NormalizedGameState option = action.getNewGamestate().normalize(killtime);
				if (!expanded.contains(option) && !border.contains(option)) {
					//we found a fresh new state
					distancetostart.put(option, distancetostart.get(expand)+action.getCost());
					if (!predecessoraction.containsKey(option)) {
						predecessoraction.put(option, new Vector<PlayerAction>());
					}
					predecessoraction.get(option).add(action);
					
					//add option to border, but at the correct position:
					int i = Collections.binarySearch(border, option, comp);
					if (i<0) i=-(i+1);else throw new Error();
					border.add(i,option);
					
				} else if (!expanded.contains(option) && border.contains(option)) {
					double distancetostart_ = distancetostart.get(expand)+action.getCost();
					if (distancetostart_<distancetostart.get(option)) {
						distancetostart.put(option, distancetostart_);
						
						//sort border again:
						border.remove(option);
						int i = Collections.binarySearch(border, option, comp);
						if (i<0) i=-(i+1);else throw new Error();
						border.add(i,option);
						
						predecessoraction.get(option).add(0,action);
					} else {
						predecessoraction.get(option).add(action);
					}
				} else {
					if (!predecessoraction.containsKey(option)) {
						predecessoraction.put(option, new Vector<PlayerAction>());
					}
					predecessoraction.get(option).add(action);
				}
			}
			expanded.add(expand);
			border.remove(expand);
		}
		//expanded now contains all NormalizedGameStates.
		{
			//check if every state can be completed to a winning state.
			HashSet<NormalizedGameState> backexpanded = new HashSet<NormalizedGameState>();
			HashSet<NormalizedGameState> backborder = new HashSet<NormalizedGameState>();
			backborder.addAll(winningstates);
			while(!backborder.isEmpty()) {
				if (killtime!=null && TimerThread.currentTimeMillis()>=killtime) throw new TimeUpException();
				NormalizedGameState backexpand = backborder.iterator().next();
				backborder.remove(backexpand);
				backexpanded.add(backexpand);
				if (!predecessoraction.containsKey(backexpand)) continue;//first game state
				for (PlayerAction action : predecessoraction.get(backexpand)) {
					NormalizedGameState pred = action.getOldGamestate().normalize(killtime);
					if (!backexpanded.contains(pred)) backborder.add(pred);
				}
			}
			expanded.removeAll(backexpanded);
			if (!expanded.isEmpty()) {
				//this game level contains deadlocks and hence we treat it as unsolvable.
				return null;
			}
		}
		return solution;
	}

}
